package expression_test

import (
	"fmt"
	"testing"

	"git.akyoto.dev/cli/q/src/build/expression"
	"git.akyoto.dev/cli/q/src/build/token"
	"git.akyoto.dev/go/assert"
)

func TestParse(t *testing.T) {
	tests := []struct {
		Name       string
		Expression string
		Result     string
	}{
		{"Identity", "1", "1"},
		{"Basic calculation", "1+2", "(+ 1 2)"},
		{"Same operator", "1+2+3", "(+ (+ 1 2) 3)"},
		{"Same operator 2", "1+2+3+4", "(+ (+ (+ 1 2) 3) 4)"},
		{"Different operator", "1+2-3", "(- (+ 1 2) 3)"},
		{"Different operator 2", "1+2-3+4", "(+ (- (+ 1 2) 3) 4)"},
		{"Different operator 3", "1+2-3+4-5", "(- (+ (- (+ 1 2) 3) 4) 5)"},
		{"Grouped identity", "(1)", "1"},
		{"Grouped identity 2", "((1))", "1"},
		{"Grouped identity 3", "(((1)))", "1"},
		{"Adding identity", "(1)+(2)", "(+ 1 2)"},
		{"Adding identity 2", "(1)+(2)+(3)", "(+ (+ 1 2) 3)"},
		{"Adding identity 3", "(1)+(2)+(3)+(4)", "(+ (+ (+ 1 2) 3) 4)"},
		{"Grouping", "(1+2)", "(+ 1 2)"},
		{"Grouping 2", "(1+2+3)", "(+ (+ 1 2) 3)"},
		{"Grouping 3", "((1)+(2)+(3))", "(+ (+ 1 2) 3)"},
		{"Grouping left", "(1+2)*3", "(* (+ 1 2) 3)"},
		{"Grouping right", "1*(2+3)", "(* 1 (+ 2 3))"},
		{"Grouping same operator", "1+(2+3)", "(+ 1 (+ 2 3))"},
		{"Grouping same operator 2", "1+(2+3)+(4+5)", "(+ (+ 1 (+ 2 3)) (+ 4 5))"},
		{"Two groups", "(1+2)*(3+4)", "(* (+ 1 2) (+ 3 4))"},
		{"Two groups 2", "(1+2-3)*(3+4-5)", "(* (- (+ 1 2) 3) (- (+ 3 4) 5))"},
		{"Two groups 3", "(1+2)*(3+4-5)", "(* (+ 1 2) (- (+ 3 4) 5))"},
		{"Operator priority", "1+2*3", "(+ 1 (* 2 3))"},
		{"Operator priority 2", "1*2+3", "(+ (* 1 2) 3)"},
		{"Operator priority 3", "1+2*3+4", "(+ (+ 1 (* 2 3)) 4)"},
		{"Operator priority 4", "1+2*(3+4)+5", "(+ (+ 1 (* 2 (+ 3 4))) 5)"},
		{"Operator priority 5", "1+2*3*4", "(+ 1 (* (* 2 3) 4))"},
		{"Operator priority 6", "1+2*3+4*5", "(+ (+ 1 (* 2 3)) (* 4 5))"},
		{"Operator priority 7", "1+2*3*4*5*6", "(+ 1 (* (* (* (* 2 3) 4) 5) 6))"},
		{"Operator priority 8", "1*2*3+4*5*6", "(+ (* (* 1 2) 3) (* (* 4 5) 6))"},
		{"Complex", "(1+2-3*4)*(5+6-7*8)", "(* (- (+ 1 2) (* 3 4)) (- (+ 5 6) (* 7 8)))"},
		{"Complex 2", "(1+2*3-4)*(5+6*7-8)", "(* (- (+ 1 (* 2 3)) 4) (- (+ 5 (* 6 7)) 8))"},
		{"Complex 3", "(1+2*3-4)*(5+6*7-8)+9-10*11", "(- (+ (* (- (+ 1 (* 2 3)) 4) (- (+ 5 (* 6 7)) 8)) 9) (* 10 11))"},
		{"Unary", "!", "!"},
		{"Function calls", "a()", "(λ a)"},
		{"Function calls 2", "a(1)", "(λ a 1)"},
		{"Function calls 3", "a(1)+1", "(+ (λ a 1) 1)"},
		{"Function calls 4", "1+a(1)", "(+ 1 (λ a 1))"},
		{"Function calls 5", "a(1,2)", "(λ a 1 2)"},
		{"Function calls 6", "a(1,2,3)", "(λ a 1 2 3)"},
		{"Function calls 7", "a(1,2+2,3)", "(λ a 1 (+ 2 2) 3)"},
		{"Function calls 8", "a(1,2+2,3+3)", "(λ a 1 (+ 2 2) (+ 3 3))"},
		{"Function calls 9", "a(1+1,2,3)", "(λ a (+ 1 1) 2 3)"},
		{"Function calls 10", "a(1+1,2+2,3+3)", "(λ a (+ 1 1) (+ 2 2) (+ 3 3))"},
		{"Function calls 11", "a(b())", "(λ a (λ b))"},
		{"Function calls 12", "a(b(),c())", "(λ a (λ b) (λ c))"},
		{"Function calls 13", "a(b(),c(),d())", "(λ a (λ b) (λ c) (λ d))"},
		{"Function calls 14", "a(b(1))", "(λ a (λ b 1))"},
		{"Function calls 15", "a(b(1),c(2),d(3))", "(λ a (λ b 1) (λ c 2) (λ d 3))"},
		{"Function calls 16", "a(b(1)+1)", "(λ a (+ (λ b 1) 1))"},
		{"Function calls 17", "a(b(1)+1,c(2),d(3))", "(λ a (+ (λ b 1) 1) (λ c 2) (λ d 3))"},
		{"Function calls 18", "a(b(1)*c(2))", "(λ a (* (λ b 1) (λ c 2)))"},
		{"Function calls 19", "a(b(1)*c(2),d(3)+e(4),f(5)/f(6))", "(λ a (* (λ b 1) (λ c 2)) (+ (λ d 3) (λ e 4)) (/ (λ f 5) (λ f 6)))"},
		{"Function calls 20", "a(b(1,2)+c(3,4)*d(5,6))", "(λ a (+ (λ b 1 2) (* (λ c 3 4) (λ d 5 6))))"},
		{"Function calls 21", "a((b(1,2)+c(3,4))*d(5,6))", "(λ a (* (+ (λ b 1 2) (λ c 3 4)) (λ d 5 6)))"},
		{"Function calls 22", "a((b(1,2)+c(3,4))*d(5,6),e())", "(λ a (* (+ (λ b 1 2) (λ c 3 4)) (λ d 5 6)) (λ e))"},
		{"Function calls 23", "a((b(1,2)+c(3,4))*d(5,6),e(7+8,9-10*11,12))", "(λ a (* (+ (λ b 1 2) (λ c 3 4)) (λ d 5 6)) (λ e (+ 7 8) (- 9 (* 10 11)) 12))"},
		{"Function calls 24", "a((b(1,2,bb())+c(3,4,cc(0)))*d(5,6,dd(0)),e(7+8,9-10*11,12,ee(0)))", "(λ a (* (+ (λ b 1 2 (λ bb)) (λ c 3 4 (λ cc 0))) (λ d 5 6 (λ dd 0))) (λ e (+ 7 8) (- 9 (* 10 11)) 12 (λ ee 0)))"},
		{"Function calls 25", "a(1-2*3)", "(λ a (- 1 (* 2 3)))"},
		{"Function calls 26", "1+2*a()+4", "(+ (+ 1 (* 2 (λ a))) 4)"},
		{"Function calls 27", "sum(a,b)*2+15*4", "(+ (* (λ sum a b) 2) (* 15 4))"},
		{"Package function calls", "math.sum(a,b)", "(λ (. math sum) a b)"},
		{"Package function calls 2", "generic.math.sum(a,b)", "(λ (. (. generic math) sum) a b)"},
		{"Array access", "a[0]", "(@ a 0)"},
		{"Array access 2", "a[b+c]", "(@ a (+ b c))"},
		{"Array access 3", "a.b[c]", "(@ (. a b) c)"},
		{"Array access 4", "a.b[c+d]", "(@ (. a b) (+ c d))"},
	}

	for _, test := range tests {
		test := test

		t.Run(test.Name, func(t *testing.T) {
			src := []byte(test.Expression)
			tokens := token.Tokenize(src)
			expr := expression.Parse(tokens)
			defer expr.Reset()

			assert.NotNil(t, expr)
			assert.Equal(t, expr.String(src), test.Result)
		})
	}
}

func TestCount(t *testing.T) {
	src := []byte("(a+b-c*d)+(a*b-c+d)")
	tokens := token.Tokenize(src)
	expr := expression.Parse(tokens)
	assert.Equal(t, expr.Count(src, token.Identifier, "a"), 2)
	assert.Equal(t, expr.Count(src, token.Identifier, "b"), 2)
	assert.Equal(t, expr.Count(src, token.Identifier, "c"), 2)
	assert.Equal(t, expr.Count(src, token.Identifier, "d"), 2)
	assert.Equal(t, expr.Count(src, token.Identifier, "e"), 0)
}

func TestEachLeaf(t *testing.T) {
	src := []byte("(1+2-3*4)+(5*6-7+8)")
	tokens := token.Tokenize(src)
	expr := expression.Parse(tokens)
	leaves := []string{}

	err := expr.EachLeaf(func(leaf *expression.Expression) error {
		leaves = append(leaves, leaf.Token.Text(src))
		return nil
	})

	assert.Nil(t, err)
	assert.DeepEqual(t, leaves, []string{"1", "2", "3", "4", "5", "6", "7", "8"})

	err = expr.EachLeaf(func(leaf *expression.Expression) error {
		return fmt.Errorf("error")
	})

	assert.NotNil(t, err)
	assert.Equal(t, err.Error(), "error")
}

func TestEachParameter(t *testing.T) {
	src := []byte("1+2,3*4,5*6,7+8")
	tokens := token.Tokenize(src)
	parameters := []string{}

	err := expression.EachParameter(tokens, func(parameter token.List) error {
		expr := expression.Parse(parameter)
		parameters = append(parameters, expr.String(src))
		return nil
	})

	assert.Nil(t, err)
	assert.DeepEqual(t, parameters, []string{"(+ 1 2)", "(* 3 4)", "(* 5 6)", "(+ 7 8)"})

	err = expression.EachParameter(tokens, func(parameter token.List) error {
		return fmt.Errorf("error")
	})

	assert.NotNil(t, err)
	assert.Equal(t, err.Error(), "error")
}

func TestRemoveChild(t *testing.T) {
	src := []byte("(1+2-3*4)+(5*6-7+8)")
	tokens := token.Tokenize(src)
	expr := expression.Parse(tokens)
	left := expr.Children[0]
	right := expr.Children[1]
	expr.RemoveChild(left)
	assert.Equal(t, expr.Children[0], right)
}

func TestNilExpression(t *testing.T) {
	src := []byte("")
	tokens := token.Tokenize(src)
	expr := expression.Parse(tokens)
	assert.Nil(t, expr)
}

func TestNilGroup(t *testing.T) {
	src := []byte("()")
	tokens := token.Tokenize(src)
	expr := expression.Parse(tokens)
	assert.Nil(t, expr)
}
